from __future__ import annotations

from hikariwave.constants import Audio
from hikariwave.audio.source import AudioSource, FileAudioSource
from typing import TYPE_CHECKING

import asyncio
import logging
import nacl.secret as secret
import struct
import time

if TYPE_CHECKING:
    from hikariwave.connection import VoiceConnection

logger: logging.Logger = logging.getLogger("hikariwave.player")

class AudioPlayer:
    """Responsible for all audio."""

    def __init__(self, connection: VoiceConnection) -> None:
        """
        Create a new audio player.
        
        Parameters
        ----------
        connection : VoiceConnection
            The active voice connection.
        """
        
        self.connection: VoiceConnection = connection

        self.ended: asyncio.Event = asyncio.Event()
        self.resumed: asyncio.Event = asyncio.Event()
        self.resumed.set()

        self.sequence: int = 0
        self.timestamp: int = 0
        self.nonce: int = 0

        self.play_task: asyncio.Task = None
    
    def _encrypt_aead_xchacha20_poly1305_rtpsize(self, header: bytes, audio: bytes) -> bytes:
        box: secret.Aead = secret.Aead(self.connection.secret)

        nonce: bytearray = bytearray(24)
        nonce[:4] = struct.pack(">I", self.nonce)

        self.nonce = (self.nonce + 1) % Audio.BIT_32U

        return header + box.encrypt(audio, header, bytes(nonce)).ciphertext + nonce[:4]

    def _generate_rtp(self) -> bytes:
        header: bytearray = bytearray(12)
        header[0] = 0x80
        header[1] = 0x78
        struct.pack_into(">H", header, 2, self.sequence)
        struct.pack_into(">I", header, 4, self.timestamp)
        struct.pack_into(">I", header, 8, self.connection.ssrc)

        return bytes(header)

    async def _play_internal(self, source: AudioSource) -> None:
        self.ended.clear()

        try:
            await self.connection.gateway.set_speaking(True)
    
            if isinstance(source, FileAudioSource):
                await self.connection.client.ffmpeg.start(source.filepath)
            else:
                await self.connection.client.ffmpeg.start(await source.read())
            
            frame_duration: float = Audio.FRAME_LENGTH / 1000
            frame_count: int = 0
            start_time: float = time.perf_counter()

            while not self.ended.is_set():
                if not self.resumed.is_set():
                    await self._send_silence()
                    await self.resumed.wait()

                    frame_count = 0
                    start_time = time.perf_counter()
                    continue

                pcm: bytes = await self.connection.client.ffmpeg.decode(Audio.FRAME_SIZE)

                if not pcm or len(pcm) < Audio.FRAME_SIZE:
                    break

                opus: bytes = await self.connection.client.opus.encode(pcm)

                if not opus:
                    break

                header: bytes = self._generate_rtp()
                encrypted: bytes = getattr(self, f"_encrypt_{self.connection.mode}")(header, opus)

                await self.connection.server.send(encrypted)

                self.sequence = (self.sequence + 1) % Audio.BIT_16U
                self.timestamp = (self.timestamp + Audio.SAMPLES_PER_FRAME) % Audio.BIT_32U
                frame_count += 1

                next_frame_time: float = start_time + (frame_count * frame_duration)
                current_time: float = time.perf_counter()
                sleep_time: float = next_frame_time - current_time

                if sleep_time > 0:
                    await asyncio.sleep(sleep_time)
                elif sleep_time < -0.020:
                    logger.debug(f"Frame {frame_count} is {-sleep_time:.3f}s behind schedule")
        finally:
            try:
                await self._send_silence()
            except:...

            try:
                await self.connection.gateway.set_speaking(False)
            except:...

    async def _send_silence(self) -> None:
        for _ in range(5): await self.connection.server.send(b"\xF8\xFF\xFE")

    async def pause(self) -> None:
        """
        Pause the current audio.
        """

        self.resumed.clear()
        await self.connection.gateway.set_speaking(False)

    async def play(self, source: AudioSource) -> None:
        """
        Play audio from a source.
        
        Parameters
        ----------
        source : AudioSource
            The source of the audio to play.
        """
        
        if self.play_task and not self.play_task.done():
            await self.stop()

            try:
                await asyncio.wait_for(self.play_task, 1)
            except:...

            await asyncio.sleep(0.05)
        
        self.play_task = asyncio.create_task(self._play_internal(source))
        await self.play_task
    
    async def resume(self) -> None:
        """
        Resume the current audio.
        """
        
        await self.connection.gateway.set_speaking(True)
        self.resumed.set()

    async def stop(self) -> None:
        """
        Stop the current audio.
        """
        
        self.ended.set()
        self.resumed.set()

        await self.connection.gateway.set_speaking(False)


from __future__ import annotations

from typing import Final, Sequence

__all__: Final[Sequence[str]] = (
    "GatewayError",
    "ServerError",
)

class GatewayError(Exception):
    """Raised when an error occurs with a voice system gateway."""

class ServerError(Exception):
    """Raised when an error occurs with a voice system server."""


from __future__ import annotations

from hikariwave.constants import Audio

import os
import sys

script_dir: str = os.path.dirname(os.path.abspath(__file__))
bin_dir: str = os.path.join(script_dir, "bin")

if sys.platform == "win32":
    os.environ["PATH"] = f"{bin_dir};{os.environ['PATH']}"
    os.add_dll_directory(bin_dir)

import opuslib

class OpusEncoder:
    """Responsible for encoding PCM audio to Opus."""

    def __init__(self) -> None:
        """
        Create an Opus encoder.
        """
        
        self.encoder: opuslib.Encoder = None
    
    async def encode(self, pcm: bytes) -> bytes:
        """
        Encode PCM into Opus audio.
        
        Parameters
        ----------
        pcm : bytes
            The PCM audio to encode.
        
        Returns
        -------
        bytes | None
            The Opus-encoded audio, if any.
        """
        
        if not self.encoder:
            return
        
        pcm = pcm[:Audio.FRAME_SIZE].ljust(
            Audio.FRAME_SIZE, b"\0",
        )
        return self.encoder.encode(pcm, Audio.SAMPLES_PER_FRAME)

    async def start(self) -> None:
        """
        Start the internal encoder to encode PCM.
        """
        
        if self.encoder: return

        self.frame_samples = (
            Audio.FRAME_SIZE //
            (Audio.CHANNELS * 2)
        )

        self.encoder = opuslib.Encoder(
            Audio.SAMPLING_RATE,
            Audio.CHANNELS,
            opuslib.APPLICATION_AUDIO,
        )
    
    async def stop(self) -> None:
        """
        Stop the internal encoder.
        """
        
        self.encoder = None


from __future__ import annotations

from hikariwave.audio.player import AudioPlayer
from hikariwave.audio.source import FileAudioSource
from hikariwave.event.types import WaveEventType
from hikariwave.gateway import Opcode, ReadyPayload, SessionDescriptionPayload, VoiceGateway
from hikariwave.server import VoiceServer
from typing import Final, Sequence, TYPE_CHECKING

import asyncio
import hikari

if TYPE_CHECKING:
    from .client import VoiceClient

__all__: Final[Sequence[str]] = ("VoiceConnection",)

class VoiceConnection:
    """An active connection to a Discord voice channel."""

    def __init__(
        self,
        client: VoiceClient,
        guild_id: hikari.Snowflakeish,
        channel_id: hikari.Snowflakeish,
        endpoint: str,
        session_id: str,
        token: str,
    ) -> None:
        """
        Create a new voice connection.
        
        Parameters
        ----------
        client : VoiceClient
            The controlling client for all connections and state.
        guild_id : hikari.Snowflakeish
            The ID of the guild the channel is in.
        channel_id : hikari.Snowflakeish
            The ID of the channel to connect to.
        endpoint : str
            The URL of Discord's voice gateway.
        session_id : str
            The provided session ID from Discord's OAuth2 gateway.
        token : str
            The provided token from Discord's OAuth2 gateway.
        """
        
        self.client: VoiceClient = client
        self.guild_id: hikari.Snowflakeish = guild_id
        self.channel_id: hikari.Snowflakeish = channel_id
        self.endpoint: str = endpoint
        self.session_id: str = session_id
        self.token: str = token

        self.server: VoiceServer = VoiceServer(self.client)
        self.gateway: VoiceGateway = VoiceGateway(
            self,
            self.guild_id,
            self.channel_id,
            self.client.bot.get_me().id,
            self.session_id,
            self.token,
        )
        self.gateway.set_callback(Opcode.READY, self._gateway_ready)
        self.gateway.set_callback(Opcode.SESSION_DESCRIPTION, self._gateway_session_description)
        self.ready: asyncio.Event = asyncio.Event()

        self.ssrc: int = None
        self.mode: str = None

        self.player: AudioPlayer = AudioPlayer(self)
    
    async def _connect(self) -> None:
        await self.gateway.connect(f"{self.endpoint}/?v=8")
        await self.ready.wait()

    async def _disconnect(self) -> None:
        await self.server.disconnect()
        await self.gateway.disconnect()

    async def _gateway_ready(self, payload: ReadyPayload) -> None:
        self.ssrc = payload.ssrc
        
        supported_mode: str = "aead_xchacha20_poly1305_rtpsize"
        ip, port = await self.server.connect(payload.ip, payload.port, self.ssrc)

        await self.gateway.select_protocol(ip, port, supported_mode)

    async def _gateway_reconnect(self) -> None:
        self.gateway = VoiceGateway(
            self,
            self.guild_id,
            self.channel_id,
            self.client.bot.get_me().id,
            self.session_id,
            self.token,
        )
        await self.gateway.connect(f"{self.endpoint}/?v=8")

        self.client.event_factory.emit(
            WaveEventType.VOICE_RECONNECT,
            self.channel_id,
            self.guild_id,
        )

    async def _gateway_session_description(self, payload: SessionDescriptionPayload) -> None:
        self.mode = payload.mode
        self.secret = payload.secret
        self.ready.set()
    
    async def disconnect(self) -> None:
        """
        Disconnect from the current channel.
        """
        
        await self.client.disconnect(self.guild_id)
    
    @property
    def latency_gateway(self) -> float:
        """Get the heartbeat latency of this connection with Discord's gateway."""
        
        return self.gateway.last_heartbeat_ack - self.gateway.last_heartbeat_sent
    
    async def pause(self) -> None:
        """
        Pause playback of the current audio.
        """
        
        await self.player.pause()

    async def play_file(self, filepath: str) -> None:
        """
        Play audio from a file.
        
        Parameters
        ----------
        filepath : str
            The path, absolute or relative, to the file to play.
        """
        
        source: FileAudioSource = FileAudioSource(filepath)
        await self.player.play(source)
    
    async def resume(self) -> None:
        """
        Resume playback of the current audio.
        """
        
        await self.player.resume()
    
    async def stop(self) -> None:
        """
        Stop playback of the current audio.
        """
        
        await self.player.stop()


from __future__ import annotations

from hikariwave.error import ServerError
from hikariwave.event.types import WaveEventType
from typing import Callable, Final, Sequence, TYPE_CHECKING

if TYPE_CHECKING:
    from .client import VoiceClient

import asyncio
import hikari
import logging
import struct

__all__: Final[Sequence[str]] = ("VoiceServer",)

logger: logging.Logger = logging.getLogger("hikari-wave.server")

class Protocol(asyncio.DatagramProtocol):
    """Background UDP protocol for communication with Discord's voice servers."""

    def __init__(self, ip_discover_future: asyncio.Future[bytes], rtp_listener: Callable[[int], None]) -> None:
        """
        Create a UDP server protocol.
        
        Parameters
        ----------
        ip_discover_future : asyncio.Future[bytes]
            The future to set when our IP is discovered.
        rtp_listener : Callable[[int], None]
            The callback to call when we receive a non-IP discovery packet.
        """
        
        self.transport: asyncio.DatagramTransport = None

        self.ip_discover_future: asyncio.Future[bytes] = ip_discover_future
        self.rtp_listener: Callable[[int], None] = rtp_listener

    def connection_made(self, transport: asyncio.DatagramTransport) -> None:
        """
        Called automatically when a UDP connection is made.
        
        Parameters
        ----------
        transport : asyncio.DatagramTransport
            The UDP transport.
        """
        
        self.transport = transport

    def datagram_received(self, data: bytes, address: tuple[str, int]) -> None:
        """
        Automatically called when we receive a UDP packet.
        
        Parameters
        ----------
        data : bytes
            The UDP packet data
        address : tuple[str, int]
            The address that sent this packet.
        """

        if not self.ip_discover_future.done():
            self.ip_discover_future.set_result(data)
            return

        self.rtp_listener(data)
    
    def error_received(self, exc: Exception):
        """
        Automatically called when an error occurs.
        
        Parameters
        ----------
        exc : Exception
            The error that occurred.
        """
        
        if self.ip_discover_future.done():
            self.ip_discover_future.set_exception(exc)
            return 

class VoiceServer:
    """The background server responsible for communicating with Discord's voice servers."""

    def __init__(
        self,
        client: VoiceClient,
    ) -> None:
        """
        Create a new voice server connection.
        
        Parameters
        ----------
        client : VoiceClient
            The voice client handling all bot connections and state.
        """
        
        self.client: VoiceClient = client

        self.ip: str = None
        self.port: int = None
        self.ssrc: int = None
        self.udp: asyncio.DatagramTransport = None

        self._last_audio: dict[int, float] = {}
        self._watch_task: asyncio.Task = None

    async def _discover_ip(self) -> tuple[str, int]:
        loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
        future: asyncio.Future[bytes] = loop.create_future()

        self.udp, _ = await loop.create_datagram_endpoint(
            lambda: Protocol(future, self._rtp_packet),
            remote_addr=(self.ip, self.port),
        )

        packet: bytes = struct.pack("!HHI", 0x1, 70, self.ssrc) + bytes(70)
        self.udp.sendto(packet)

        data: bytes = await asyncio.wait_for(future, 3.0)

        if (type_ := struct.unpack("!H", data[0:2])[0]) != 0x0002:
            error: str = f"Expected packet type 2, got {type_}"
            raise ServerError(error)
    
        if (len_ := struct.unpack("!H", data[2:4])[0]) != 70:
            error: str = f"Expected packet length of 70, got {len_}"
            raise ServerError(error)

        external_ip: str = data[8:72].split(b"\x00", 1)[0].decode("ascii")
        external_port: int = struct.unpack("!H", data[72:74])[0]

        logger.debug(f"External address discovered: ({external_ip}:{external_port})")
        return external_ip, external_port

    def _rtp_packet(self, data: bytes) -> None:
        if len(data) < 12: return

        ssrc: int = struct.unpack_from(">I", data, 8)[0]

        if ssrc not in self.client.ssrcs_reference:
            return

        now: float = asyncio.get_running_loop().time()
        is_new: bool = ssrc not in self._last_audio
        self._last_audio[ssrc] = now

        if not is_new:
            return
        
        user_id: hikari.Snowflake = self.client.ssrcs_reference[ssrc]
        member: hikari.Member = self.client.views[user_id]
        channel_id: hikari.Snowflake = self.client.members[member.id].channel_id
        channel = self.client.channels[channel_id]
        guild: hikari.Snowflake = channel.guild_id
        self.client.event_factory.emit(
            WaveEventType.MEMBER_START_SPEAKING,
            channel_id,
            guild,
            member,
        )

        if len(channel.active) == 0:
            self.client.event_factory.emit(
                WaveEventType.VOICE_ACTIVE,
                channel_id,
                guild,
            )
        
        channel.active.add(user_id)

    async def _watch_silence(self) -> None:
        loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()

        try:
            while True:
                now: float = loop.time()

                for ssrc, last in list(self._last_audio.items()):
                    if now - last <= 0.25:
                        continue

                    del self._last_audio[ssrc]

                    user: hikari.Snowflake = self.client.ssrcs_reference[ssrc]
                    member: hikari.Member = self.client.views[user]
                    channel_id: hikari.Snowflake = self.client.members[member.id].channel_id
                    channel = self.client.channels[channel_id]
                    guild: hikari.Snowflake = channel.guild_id

                    channel.active.remove(user)

                    self.client.event_factory.emit(
                        WaveEventType.MEMBER_STOP_SPEAKING,
                        channel,
                        guild,
                        member,
                    )
                
                    if len(channel.active) == 0:
                        self.client.event_factory.emit(
                            WaveEventType.VOICE_INACTIVE,
                            channel_id,
                            guild,
                        )

                await asyncio.sleep(0.05)
        except asyncio.CancelledError:
            return

    async def connect(self, ip: str, port: int, ssrc: int) -> tuple[str, int]:
        """
        Connect to a Discord voice server.
        
        Parameters
        ----------
        ip : str
            The Discord voice server's IP.
        port : int
            The Discord voice server's port.
        ssrc : int
            Our assigned SSRC by Discord's voice gateway.
        
        Returns
        -------
        tuple[str, int]
            Our public, discovered IP address.
        """
        
        self.ip, self.port, self.ssrc = ip, port, ssrc

        logger.debug(f"Connecting to voice server: IP={ip}, Port={port}, SSRC={ssrc}")
        local_address: tuple[str, int] = await self._discover_ip()

        self._watch_task = asyncio.create_task(self._watch_silence())

        return local_address

    async def disconnect(self) -> None:
        """
        Disconnect from Discord's voice server.
        """
        
        logger.debug(f"Disconnected from server: IP={self.ip}, Port={self.port}")
        
        if self._watch_task:
            self._watch_task.cancel()
            self._watch_task = None
        
        self._last_audio.clear()

        if self.udp:
            self.udp.close()
            self.udp = None
    
    async def send(self, data: bytes) -> None:
        """
        Send a UDP packet to Discord's voice server.
        
        Parameters
        ----------
        data : bytes
            The UDP packet to send.
        """
        
        if not self.udp or self.udp.is_closing():
            return
    
        self.udp.sendto(data)


from __future__ import annotations

from enum import IntEnum
from typing import Final, Sequence

import io
import struct

__all__: Final[Sequence[str]] = (
    "Audio",
    "CloseCode",
    "Opcode",
)

class Audio:
    """Collection of audio-related constants."""

    BIT_16U: int = 2**16 - 1
    BIT_32U: int = 2**32 - 1
    BLOCKSIZE: int = io.DEFAULT_BUFFER_SIZE
    CHANNELS: int = 2
    FRAME_LENGTH: int = 20
    SAMPLE_SIZE: int = struct.calcsize('h') * CHANNELS
    SAMPLING_RATE: int = 48000
    SAMPLES_PER_FRAME: int = int(SAMPLING_RATE / 1000 * FRAME_LENGTH)

    FRAME_SIZE: int = SAMPLES_PER_FRAME * SAMPLE_SIZE

class CloseCode(IntEnum):
    """Collection of a voice close event codes."""

    UNKNOWN_OPCODE = 4001
    """Client sent an invalid opcode."""
    FAILED_TO_DECODE_PAYLOAD = 4002
    """Client send an invalid payload during IDENTIFY."""
    NOT_AUTHENTICATED = 4003
    """Client sent a payload before IDENTIFY."""
    AUTHENTICATION_FAILED = 4004
    """Client sent an incorrect token in IDENTIFY."""
    ALREADY_AUTHENTICATED = 4005
    """Client sent more than one IDENTIFY."""
    SESSION_NO_LONGER_VALID = 4006
    """Client session is not longer valid. Reconnection required."""
    SESSION_TIMEOUT = 4009
    """Client session timed out. Reconnection required."""
    SERVER_NOT_FOUND = 4011
    """Client attempted to connect to a server that wasn't found."""
    UNKNOWN_PROTOCOL = 4012
    """Client sent a protocol that is unrecognized by the server."""
    DISCONNECTED = 4014
    """Client was disconnected. No reconnection."""
    VOICE_SERVER_CRASHED = 4015
    """Server crashed. Resume required."""
    UNKNOWN_ENCRYPTION_MODE = 4016
    """Client sent an unrecognized encryption method."""
    BAD_REQUEST = 4020
    """Client send a malformed request."""
    DISCONNECTED_RATE_LIMITED = 4021
    """Client was disconnected due to rate limit being exceeded. No reconnection."""
    DISCONNECTED_CALL_TERMINATED = 4022
    """Client was disconnected due to call being terminated. No reconnection."""

class Opcode(IntEnum):
    """Collection of voice gateway operation codes."""

    IDENTIFY = 0
    """`CLIENT` - Begin a voice websocket connection."""
    SELECT_PROTOCOL = 1
    """`CLIENT` - Select the voice protocol."""
    READY = 2
    """`SERVER` - Complete the websocket handshake."""
    HEARTBEAT = 3
    """`CLIENT` - Keep the websocket connection alive."""
    SESSION_DESCRIPTION = 4
    """`SERVER` - Describe the session."""
    SPEAKING = 5
    """`CLIENT/SERVER` - Indicate which users are speaking."""
    HEARTBEAT_ACK = 6
    """`SERVER` - Sent to acknowledge a received client heartbeat."""
    RESUME = 7
    """`CLIENT` - Resume a connection."""
    HELLO = 8
    """`SERVER` - Time to wait between sending heartbeats in milliseconds."""
    RESUMED = 9
    """`SERVER` - Acknowledge a successful session resume."""
    CLIENTS_CONNECT = 11
    """`SERVER` - One or more clients have connection to the voice channel."""
    CLIENT_DISCONNECT = 13
    """`SERVER` - A client has disconnected from the voice channel."""
    UNDOCUMENTED_15 = 15
    """`SERVER` - Unknown as of `0.0.1a1`."""
    UNDOCUMENTED_18 = 18
    """`SERVER` - Unknown as of `0.0.1a1` - Contains user flags of user who joined channel."""
    UNDOCUMENTED_20 = 20
    """`SERVER` - Unknown as of `0.0.1a1` - Contains user platform of user who joined channel."""


from __future__ import annotations

from dataclasses import dataclass
from hikariwave.audio.ffmpeg import FFmpegDecoder
from hikariwave.audio.opus import OpusEncoder
from hikariwave.connection import VoiceConnection
from hikariwave.error import GatewayError
from hikariwave.event.factory import EventFactory
from hikariwave.event.types import WaveEventType
from typing import Final, Sequence, TypeAlias

import asyncio
import hikari
import logging

__all__: Final[Sequence[str]] = ("VoiceClient",)

logger: logging.Logger = logging.getLogger("hikari-wave.client")

ChannelID: TypeAlias = hikari.Snowflakeish
GuildID: TypeAlias = hikari.Snowflakeish
MemberID: TypeAlias = hikari.Snowflakeish

@dataclass(slots=True)
class VoiceChannelMeta:
    """Metadata container for a voice channel."""

    active: set[hikari.Snowflake]
    """A set of all actively speaking users in this channel."""
    guild_id: hikari.Snowflake
    """The ID of the guild this channel is in."""
    id: hikari.Snowflake
    """The ID of this channel."""
    population: int
    """How many users are in this channel."""

@dataclass(slots=True)
class VoiceMemberMeta:
    """Metadata container for a member."""

    channel_id: hikari.Snowflake
    """The ID of the channel this member is in."""
    is_mute: bool
    """If the member is muted."""
    is_deaf: bool
    """If the member is deafened."""

class VoiceClient:
    """Voice system implementation for `hikari`-based Discord bots."""

    def __init__(
        self,
        bot: hikari.GatewayBot,
    ) -> None:
        """
        Create a new voice client.
        
        Parameters
        ----------
        bot : hikari.GatewayBot
            The `hikari`-based Discord bot to link this voice system with.
        """
        
        self.bot: hikari.GatewayBot = bot
        self.bot.subscribe(hikari.VoiceStateUpdateEvent, self._disconnected)
        self.bot.subscribe(hikari.VoiceStateUpdateEvent, self._voice_state_update)

        self.connections: dict[GuildID, VoiceConnection] = {}
        self.connections_reference: dict[ChannelID, GuildID] = {}

        self.members: dict[MemberID, VoiceMemberMeta] = {}
        self.channels: dict[ChannelID, VoiceChannelMeta] = {}
        self.views: dict[MemberID, hikari.Member] = {}
        self.ssrcs: dict[MemberID, int] = {}
        self.ssrcs_reference: dict[int, MemberID] = {}

        self.event_factory: EventFactory = EventFactory(self.bot)

        self.ffmpeg: FFmpegDecoder = FFmpegDecoder()
        self.opus: OpusEncoder = OpusEncoder()
    
    async def _disconnect(self, guild_id: hikari.Snowflakeish) -> None:
        connection: VoiceConnection = self.connections.pop(guild_id)

        logger.info(f"Disconnecting from voice: Guild={guild_id}, Channel={connection.channel_id}")

        del self.connections_reference[connection.channel_id]
        await connection._disconnect()

        self.event_factory.emit(
            WaveEventType.BOT_LEAVE_VOICE,
            self.bot,
            connection.channel_id,
            guild_id,
        )

        if len(self.connections) == 0:
            await self.opus.stop()
            await self.ffmpeg.stop()

    async def _disconnected(self, event: hikari.VoiceStateUpdateEvent) -> None:
        if event.state.user_id != self.bot.get_me().id:
            return
        
        if event.guild_id not in self.connections:
            return
        
        await self._disconnect(event.guild_id)

    async def _voice_state_update(self, event: hikari.VoiceStateUpdateEvent) -> None:
        state: hikari.VoiceState = event.state

        if state.user_id == self.bot.get_me().id: return
        if not state.member: return

        channel: hikari.Snowflake = state.channel_id
        guild: hikari.Snowflake = state.guild_id
        member: hikari.Member = state.member

        in_voice: bool = member.id in self.members

        if channel and not in_voice:
            self.members[member.id] = VoiceMemberMeta(channel, state.is_guild_muted or state.is_self_muted, state.is_guild_deafened or state.is_self_deafened)
            self.views[member.id] = member

            if channel not in self.channels:
                self.channels[channel] = VoiceChannelMeta(set(), guild, channel, 0)
                self.event_factory.emit(
                    WaveEventType.VOICE_POPULATED,
                    channel,
                    guild,
                )
            
            self.channels[channel].population += 1
            self.event_factory.emit(
                WaveEventType.MEMBER_JOIN_VOICE,
                channel,
                guild,
                member,
            )
            return
        
        if not channel and in_voice:
            channel = self.members[member.id].channel_id
            cmeta: VoiceChannelMeta = self.channels[channel]
            cmeta.population -= 1

            if cmeta.population < 1:
                del self.channels[channel]
                self.event_factory.emit(
                    WaveEventType.VOICE_EMPTY,
                    channel,
                    guild,
                )

            del self.members[member.id]
            del self.views[member.id]

            if member.id in self.ssrcs:
                ssrc: int = self.ssrcs.pop(member.id)
                del self.ssrcs_reference[ssrc]
            
            return self.event_factory.emit(
                WaveEventType.MEMBER_LEAVE_VOICE,
                channel,
                guild,
                member,
            )
        
        if channel and in_voice and self.members[member.id].channel_id != channel:
            meta: VoiceMemberMeta = self.members[member.id]
            self.event_factory.emit(
                WaveEventType.MEMBER_MOVE_VOICE,
                guild,
                member,
                channel,
                meta.channel_id,
            )
            meta.channel_id = channel

            self.views[member.id] = member
            return

        if channel:
            meta: VoiceMemberMeta = self.members[member.id]
            deaf: bool = event.state.is_self_deafened or event.state.is_guild_deafened
            mute: bool = event.state.is_self_muted or event.state.is_guild_muted

            if deaf != meta.is_deaf:
                meta.is_deaf = deaf
                self.event_factory.emit(
                    WaveEventType.MEMBER_DEAF,
                    channel,
                    guild,
                    member,
                    deaf,
                )
            
            if mute != meta.is_mute:
                meta.is_mute = mute
                self.event_factory.emit(
                    WaveEventType.MEMBER_MUTE,
                    channel,
                    guild,
                    member,
                    mute,
                )
            
            self.views[member.id] = member

    async def connect(
        self,
        guild_id: hikari.Snowflakeish,
        channel_id: hikari.Snowflakeish,
        *,
        mute: bool = False,
        deaf: bool = True,
    ) -> VoiceConnection:
        """
        Connect to a voice channel.
        
        Parameters
        ----------
        guild_id : hikari.Snowflakeish
            The ID of the guild that the channel is in.
        channel_id : hikari.Snowflakeish
            The ID of the channel to connect to.
        mute : bool 
            If the bot should be muted upon joining the channel - Default `False`.
        deaf : bool
            If the bot should be deafened upon joining the channel - Default `True`.
        
        Returns
        -------
        VoiceConnection
            The active connection to the voice channel, once fully connected.
        
        Raises
        ------
        asyncio.TimeoutError
            If Discord doesn't send a corresponding voice server/state update.
        """

        logger.info(f"Connecting to voice: Guild={guild_id}, Channel={channel_id}, Mute={mute}, Deaf={deaf}")

        await self.bot.update_voice_state(guild_id, channel_id, self_mute=mute, self_deaf=deaf)

        try:
            server_update, state_update = await asyncio.gather(
                self.bot.wait_for(
                    hikari.VoiceServerUpdateEvent, 3.0,
                    lambda e: e.guild_id == guild_id
                ),
                self.bot.wait_for(
                    hikari.VoiceStateUpdateEvent, 3.0,
                    lambda e: e.guild_id == guild_id and e.state.channel_id == channel_id and e.state.user_id == self.bot.get_me().id
                )
            )
        except asyncio.TimeoutError:
            error: str = "Voice server/state update timed out"
            raise GatewayError(error)

        guild: hikari.Guild = await self.bot.rest.fetch_guild(guild_id)
        population: int = 0
        for state in guild.get_voice_states().values():
            self.members[state.member.id] = VoiceMemberMeta(channel_id, state.is_guild_muted or state.is_self_muted, state.is_guild_deafened or state.is_self_deafened)
            self.views[state.member.id] = state.member
            
            population += 1
        self.channels[channel_id] = VoiceChannelMeta(set(), guild_id, channel_id, population)

        connection: VoiceConnection = VoiceConnection(
            self,
            guild_id,
            channel_id,
            server_update.endpoint,
            state_update.state.session_id,
            server_update.token,
        )
        await connection._connect()
        self.connections[guild_id] = connection
        self.connections_reference[channel_id] = guild_id

        self.event_factory.emit(
            WaveEventType.BOT_JOIN_VOICE,
            self.bot,
            channel_id,
            guild_id,
            deaf,
            mute,
        )

        if len(self.connections) == 1:
            await self.opus.start()

        return connection
    
    async def disconnect(
        self,
        *,
        guild_id: hikari.Snowflakeish | None = None,
        channel_id: hikari.Snowflakeish | None = None,
    ) -> None:
        """
        Disconnect from a voice channel.
        
        Parameters
        ----------
        guild_id : hikari.Snowflakeish | None
            The ID of the guild that the channel to disconnect from is in - Default `None`.
        channel_id : hikari.Snowflakeish | None
            The ID of the channel to disconnect from - Default `None`.
        
        Note
        ----
        At least one of `guild_id` or `channel_id` must be provided.

        Raises
        ------
        ValueError
            If neither of `guild_id` or `channel_id` are provided.
        """
        
        if not guild_id and not channel_id:
            error: str = "At least guild_id or channel_id must be defined"
            raise ValueError(error)
        
        if channel_id:
            guild_id = self.connections_reference[channel_id]

        await self.bot.update_voice_state(guild_id, None)
        await self._disconnect(guild_id)
    
    def get_connection(
        self,
        *,
        guild_id: hikari.Snowflakeish | None = None,
        channel_id: hikari.Snowflakeish | None = None,
    ) -> VoiceConnection | None:
        """
        Get an active voice connection.
        
        Parameters
        ----------
        guild_id : hikari.Snowflakeish | None
            The ID of the guild that the connection is handling - Default `None`.
        channel_id : hikari.Snowflakeish | None
            The ID of the channel that the connection is handling - Default `None`.
        
        Note
        ----
        At least one of `guild_id` or `channel_id` must be provided.
        
        Returns
        -------
        VoiceConnection | None
            The active voice connection at the guild/channel, if present.

        Raises
        ------
        ValueError
            If neither of `guild_id` or `channel_id` are provided.
        """

        if not guild_id and not channel_id:
            error: str = "At least guild_id or channel_id must be defined"
            raise ValueError(error)
        
        if channel_id:
            guild_id = self.connections_reference[channel_id]
        
        try:
            return self.connections[guild_id]
        except KeyError:
            return


from __future__ import annotations

from hikariwave.event import events
from hikariwave.event.types import WaveEventType
from typing import Callable, Final, Sequence, TypeAlias

import hikari

__all__: Final[Sequence[str]] = ("EventFactory",)

EventBuilder: TypeAlias = Callable[[tuple], events.WaveEvent]

class EventFactory:
    """Responsible for emitting and handling all supplemental events."""

    __slots__ = (
        "_bot",
        "_builders",
    )

    def __init__(self, bot: hikari.GatewayBot) -> None:
        """
        Create a new event factory.
        
        Parameters
        ----------
        bot : hikari.GatewayBot
            The OAuth2 bot to use for dispatching events.
        """
        
        self._bot: hikari.GatewayBot = bot
        self._builders: dict[WaveEventType, EventBuilder] = {}

        for type_, event in {
            WaveEventType.BOT_JOIN_VOICE: events.BotJoinVoiceEvent,
            WaveEventType.BOT_LEAVE_VOICE: events.BotLeaveVoiceEvent,
            WaveEventType.MEMBER_DEAF: events.MemberDeafEvent,
            WaveEventType.MEMBER_JOIN_VOICE: events.MemberJoinVoiceEvent,
            WaveEventType.MEMBER_LEAVE_VOICE: events.MemberLeaveVoiceEvent,
            WaveEventType.MEMBER_MOVE_VOICE: events.MemberMoveVoiceEvent,
            WaveEventType.MEMBER_MUTE: events.MemberMuteEvent,
            WaveEventType.MEMBER_START_SPEAKING: events.MemberStartSpeakingEvent,
            WaveEventType.MEMBER_STOP_SPEAKING: events.MemberStopSpeakingEvent,
            WaveEventType.VOICE_ACTIVE: events.VoiceActiveEvent,
            WaveEventType.VOICE_EMPTY: events.VoiceEmptyEvent,
            WaveEventType.VOICE_INACTIVE: events.VoiceInactiveEvent,
            WaveEventType.VOICE_POPULATED: events.VoicePopulatedEvent,
            WaveEventType.VOICE_RECONNECT: events.VoiceReconnectEvent,
            WaveEventType.VOICE_WARNING: events.VoiceWarningEvent,
        }.items():
            self._builders[type_] = event._create
    
    def emit(self, key: WaveEventType, *args) -> None:
        """
        Emit a supplemental event.
        
        Parameters
        ----------
        key : WaveEventType
            The type of supplemental event to emit.
        args : tuple
            Any arguments needed in the instantiation of such event.
        
        Raises
        ------
        KeyError
            If the provided key doesn't correspond to a supplemental event.
        """
        
        try:
            self._bot.dispatch(self._builders[key](*args))
        except KeyError:
            error: str = f"No event registered for key {key!r}"
            raise RuntimeError(error)


from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, AsyncGenerator

import aiofiles

class AudioSource(ABC):
    """Base audio source implementation."""

    @abstractmethod
    async def read(self) -> bytes:
        """
        Read the entire contents of this source.
        
        Returns
        -------
        bytes
            The contents of this source.
        """

    @abstractmethod
    async def stream(self, size: int) -> AsyncGenerator[bytes, Any]:
        """
        Stream the contents of this source in chunks.
        
        Parameters
        ----------
        size : int
            The amount of bytes to stream in each chunk.
        
        Yields
        ------
        bytes
            Each chunk of audio from this source.
        """

class FileAudioSource(AudioSource):
    """File audio source implementation."""

    def __init__(self, filepath: str) -> None:
        """
        Create a file audio source.
        
        Parameters
        ----------
        filepath : str
            The path, relative or absolute, to the audio file.
        """
        
        self.filepath: str = filepath

    async def read(self) -> bytes:
        async with aiofiles.open(self.filepath, "rb") as file:
            return await file.read()
    
    async def stream(self, size: int) -> AsyncGenerator[bytes, Any]:
        async with aiofiles.open(self.filepath, "rb") as file:
            while True:
                chunk: bytes = await file.read(size)

                if not chunk or len(chunk) != size:
                    return

                yield chunk


"""
### hikari-wave: `0.0.1a1`\n
An asynchronous, type-safe, easy-to-use voice system implementation for `hikari`-based Discord bots.

**Documentation:** https://hikari-wave.wildevstudios.net/en/0.0.1a1\n
**GitHub:** https://github.com/WilDev-Studios/hikari-wave
"""

from .client import VoiceClient
from .connection import VoiceConnection
from .error import *
from .event.events import *


from __future__ import annotations

from hikariwave.constants import Audio

import asyncio
import logging

logger: logging.Logger = logging.getLogger("hikariwave.ffmpeg")

class FFmpegDecoder:
    """Decodes audio into PCM frames."""

    def __init__(self) -> None:
        """
        Create a new FFmpeg decoder.
        """
        
        self.process: asyncio.subprocess.Process = None
    
    async def decode(self, size: int) -> bytes:
        """
        Decode the audio into PCM.

        Parameters
        ----------
        size : int
            The amount of audio to return.
        
        Returns
        -------
        bytes | None
            The decoded PCM audio.
        """

        if not self.process: return
        return await self.process.stdout.read(size)

    async def start(self, raw: bytes | str) -> None:
        """
        Start FFmpeg decoding audio.
        
        Parameters
        ----------
        raw : bytes | str
            The raw audio data or filepath.
        """

        if self.process:
            await self.stop()
        
        if isinstance(raw, bytes):
            self.process = await asyncio.create_subprocess_exec(
                "ffmpeg",
                "-i", "pipe:0",
                "-f", "s16le",
                "-ar", str(Audio.SAMPLING_RATE),
                "-ac", str(Audio.CHANNELS),
                "-loglevel", "warning",
                "-blocksize", str(Audio.BLOCKSIZE),
                "pipe:1",
                stdin=asyncio.subprocess.PIPE,
                stdout=asyncio.subprocess.PIPE,
            )

            self.process.stdin.write(raw)
            await self.process.stdin.drain()
            self.process.stdin.close()
            await self.process.stdin.wait_closed()
        else:
            self.process = await asyncio.create_subprocess_exec(
                "ffmpeg",
                "-i", raw,
                "-f", "s16le",
                "-ar", str(Audio.SAMPLING_RATE),
                "-ac", str(Audio.CHANNELS),
                "-loglevel", "warning",
                "-blocksize", str(Audio.BLOCKSIZE),
                "pipe:1",
                stdout=asyncio.subprocess.PIPE,
            )
    
    async def stop(self) -> None:
        """
        Stop the decoding of audio.
        """
        
        if not self.process:
            return
        
        try:
            if self.process.stdin and not self.process.stdin.is_closing():
                try:
                    self.process.stdin.close()
                    await self.process.stdin.wait_closed()
                except:...
            
            if self.process.stdout and not self.process.stdout.at_eof():
                try:
                    self.process.stdout.feed_eof()
                except:...
            
            if self.process.returncode is None:
                self.process.terminate()

                try:
                    await asyncio.wait_for(self.process.wait(), 0.5)
                except asyncio.TimeoutError:
                    self.process.kill()
                    await self.process.wait()
        except ProcessLookupError:...
        except Exception as e:
            logger.debug(f"Error stopping FFmpeg: {e}")
        finally:
            self.process = None